@@ -23,6 +23,11 @@
 
 using namespace std;
 
+#ifndef RECURSOR
+#include "statbag.hh"
+StatBag S;
+#endif
+
 ArgvMap &arg()
 {
   static ArgvMap arg;
